/*
CApplication class for Linux
24 October 2009
*/

#include <errno.h>
#include <stdio.h>
#include <sys/stat.h>
#include <SDL/SDL.h>
#include <SDL/SDL_thread.h>
#include <SDL/SDL_syswm.h>
#include <SDL/SDL_image.h>
#include <GL/gl.h>
#include <GL/glx.h>

#include "Shared.h"
#include "../Shared/CObeliskException.h"
#include "CApplication.h"
#include "CThreadWrapper.h"
#include "CFileWrapper.h"

using namespace std;
using namespace Obelisk;



IMainController::~IMainController()
{
}



CApplication::CApplication()
{
	mDirectorySet = false;
	mCaptionSet = false;
	mWindowXSize = -1;
	mTimerID = NULL;
	mTimerCallbackActive = false;
	mMouseLeftLast = false;
	mMouseRightLast = false;
	mAltAttached = false;
	mAudioInitialised = false;

	srand(time(NULL));
}



CApplication::~CApplication()
{
}



void CApplication::Run()
{
	mMainController = CreateMainController(this);
	Initialise();
	mMainController->Initialise();
	bool Stop = false;
	while (!Stop)
	{
		SDL_Event Event;
		SDL_WaitEvent(&Event);
		switch (Event.type)
		{
		case SDL_QUIT:
		{
			if (mTimerID != NULL)
				StopTimer();
			for (TThreadMap::const_iterator i = mThreads.begin(); i != mThreads.end(); i++)
			{
				i->second->Stop();
				delete i->second;
			}
			Stop = true;
			break;
		}
		case SDL_USEREVENT:
			switch (Event.user.code)
			{
			case 0:   // notify
				mMainController->OnNotify((int)Event.user.data1, (int)Event.user.data2);
				break;
			case 1:
				ThreadFinished((IThread*)Event.user.data1);
				break;
			case 2:
				throw CException("Thread crashed");
			case 3:
			{
				char *Message = (char*)Event.user.data1;
				CString ExceptionMessage(strlen(Message), Message);
				delete [] Message;
				throw CObeliskException(ExceptionMessage);
			}
			case 4:   // timer
				OnTimer();
				break;
			}
			break;
		case SDL_KEYDOWN:
			OnKeyDown(Event.key.keysym.sym);
			break;
		case SDL_KEYUP:
			OnKeyUp(Event.key.keysym.sym);
			break;
		case SDL_ACTIVEEVENT:
			if (Event.active.gain == 0)
			{
				if ((SDL_GetAppState() & (SDL_APPINPUTFOCUS | SDL_APPACTIVE)) !=
						(SDL_APPINPUTFOCUS | SDL_APPACTIVE))
					mMainController->OnDeactivate();
			}
			else if (Event.active.gain == 1)
			{
				if ((SDL_GetAppState() & (SDL_APPINPUTFOCUS | SDL_APPACTIVE)) ==
						(SDL_APPINPUTFOCUS | SDL_APPACTIVE))
					mMainController->OnActivate();
			}
			break;
		}
	}
	delete mMainController;
	Destroy();
}



void CApplication::Initialise()
{
	if (!mCaptionSet)
		throw CObeliskException(SL("Program forgot to call SetCaption()"));
	if (mWindowXSize == -1)
		throw CObeliskException(SL("Program forgot to call SetSize()"));

	if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER | SDL_INIT_NOPARACHUTE) != 0)
		throw CException("SDL_Init() failed");
	SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);

	mSurface = SDL_SetVideoMode(mWindowXSize, mWindowYSize, 32, SDL_SWSURFACE);
	if (mSurface == NULL)
		throw CException("SDL_SetVideoMode() failed\n");
	SDL_WM_SetCaption(mCaption.GetData(), mCaption.GetData());

	SDL_SysWMinfo Info;
	SDL_VERSION(&Info.version);
	SDL_GetWMInfo(&Info);
	mDisplay = Info.info.x11.display;
	mWindow = Info.info.x11.window;

	int Attributes[] = {
			GLX_DOUBLEBUFFER,
			GLX_RGBA, GLX_RED_SIZE, 8, GLX_GREEN_SIZE, 8, GLX_BLUE_SIZE, 8,
			GLX_DEPTH_SIZE, 16,
			None };
	XVisualInfo *VisualInfo = glXChooseVisual(mDisplay, DefaultScreen(mDisplay), Attributes);
	if (VisualInfo == NULL)
		throw CObeliskException(SL("glXChooseVisual() failed"));
	mRenderingContext = glXCreateContext(mDisplay, VisualInfo, NULL, true);
	if (mRenderingContext == NULL)
		throw CException("glXCreateContext() failed");
	mAltRenderingContext = glXCreateContext(mDisplay, VisualInfo, mRenderingContext, true);
	if (mAltRenderingContext == NULL)
		throw CException("glXCreateContext() failed (1)");
	XFree(VisualInfo);
	bool MCResult = glXMakeCurrent(mDisplay, mWindow, mRenderingContext);
	Assert(MCResult)
	Assert(glGetError() == GL_NO_ERROR)



	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glClearDepth(1.0f);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	glShadeModel(GL_SMOOTH);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	Assert(glGetError() == GL_NO_ERROR)
}



void CApplication::Destroy()
{
	bool Result = glXMakeCurrent(mDisplay, None, NULL);
	Assert(Result)
	glXDestroyContext(mDisplay, mRenderingContext);
	glXDestroyContext(mDisplay, mAltRenderingContext);

	SDL_PauseAudio(1);
	vector<SSoundInfo> Sounds = mSounds.GetValues();
	for (unsigned int i = 0; i < Sounds.size(); i++)
		SDL_FreeWAV(Sounds[i].Data);
	SDL_CloseAudio();

	SDL_Quit();
}



CString CApplication::GetDataFilePath(CString aFileName)
{
	return SL("Data/") + aFileName;
}



bool CApplication::GetUserFilePath(CString aFileName, bool aCreatePath, CString& aPath)
{
	if (!mDirectorySet)
		throw CObeliskException(SL("Program forgot to call SetDirectory()"));
	const char *HomeEnv = getenv("HOME");
	CString Path = CString(strlen(HomeEnv), HomeEnv);
	if (Path.GetChar(Path.Last()) != SL("/"))
		Path += SL("/");
	Path += mDirectory;
	if (!aCreatePath)
	{
		struct stat Info;
		int Result = stat(Path.GetData(), &Info);
		if (Result == -1)
		{
			Assert(errno == ENOENT)
			if (mkdir(Path.GetData(), 0777) == -1)
				return false;
		}
		else if (!S_ISDIR(Info.st_mode))
			return false;
	}
	aPath = Path + SL("/") + aFileName;
	return true;
}



void CApplication::SetDirectory(CString aDirectory)
{
	Assert(!mDirectorySet);   // TODO: disallow \ and /
	mDirectorySet = true;
	mDirectory = aDirectory;
}



void CApplication::SetCaption(CString aCaption)
{
	Assert(!mCaptionSet)
	mCaptionSet = true;
	mCaption = aCaption;
}



void CApplication::SetSize(int aXSize, int aYSize)
{
	Assert(mWindowXSize == -1 && aXSize > 0 && aYSize > 0)
	mWindowXSize = aXSize;
	mWindowYSize = aYSize;
}



CString CApplication::IntToString(int aNumber, int aLeading0s)
{
	bool Negative;
	if (aNumber >= 0)
		Negative = false;
	else
	{
		Negative = true;
		aNumber = -aNumber;
	}
	int Next = 10;
	int Length = 1;
	while (Next <= aNumber)
	{
		Next *= 10;
		Length++;
	}
	int NumLength = Max(Length, aLeading0s);
	int StrLength = Negative ? 1 + NumLength : NumLength;
	char *Str = new char [StrLength];
	int Offset;
	if (!Negative)
		Offset = 0;
	else
	{
		Str[0] = '-';
		Offset = 1;
	}
	for (int i = 0; i < aLeading0s - Length; i++)
		Str[Offset++] = '0';
	for (int i = 0; i < Length; i++)
	{
		Next /= 10;
		Str[Offset + i] = ((aNumber / Next) % 10) + '0';
	}
	return CString(StrLength, Str);
}



void CApplication::RunThread(IThread* aThread)
{
	CThreadWrapper *Thread = new CThreadWrapper(this, aThread);
	Thread->Run();
	mThreads[aThread] = Thread;
}



void CApplication::ThreadFinished(IThread* aThread)
{
	Assert(mThreads.count(aThread) == 1)

	mAltAttached = false;

	mMainController->ThreadFinished(aThread);
	delete mThreads[aThread];
	mThreads.erase(aThread);
}



void CApplication::NotifyWindow(int aMessage, int aData)
{
	SDL_Event Event;
	Event.type = SDL_USEREVENT;
	Event.user.code = 0;
	Event.user.data1 = (void*)aMessage;
	Event.user.data2 = (void*)aData;
	int Result = SDL_PushEvent(&Event);
	Assert(Result == 0)
}



void CApplication::StartTimer()
{
	Assert(mTimerID == NULL)
	mTimerID = SDL_AddTimer(20, TimerCallback, this);
	Assert(mTimerID != NULL)
	mTimerTick = false;
}



Uint32 CApplication::TimerCallback(Uint32 aInterval, void* aParam)
{
	CApplication *Application = (CApplication*)aParam;
	if (!Application->mTimerTick)
	{
		SDL_Event Event;
		Event.type = SDL_USEREVENT;
		Event.user.code = 4;
		int Result = SDL_PushEvent(&Event);
		Assert(Result == 0)
		Application->mTimerTick = true;
	}
	return aInterval;
}



void CApplication::OnTimer()
{
	if (mTimerID == NULL)   // ignore if timer is not running
		return;
	Uint8 MouseState = SDL_GetMouseState(NULL, NULL);
	mMouseLeft = (MouseState & SDL_BUTTON(1)) != 0;
	mMouseRight = (MouseState & SDL_BUTTON(3)) != 0;
	mTimerCallbackActive = true;
	mMainController->OnTimer();
	mTimerCallbackActive = false;
	mMouseLeftLast = mMouseLeft;
	mMouseRightLast = mMouseRight;

	mTimerTick = false;
}



void CApplication::StopTimer()
{
	Assert(mTimerID != NULL)
	SDL_bool Result = SDL_RemoveTimer(mTimerID);
	Assert(Result == SDL_TRUE)
	mTimerID = NULL;
}



unsigned int CApplication::CreateTimer()
{
	return SDL_GetTicks();
}



unsigned int CApplication::Duration(unsigned int aTimer)
{
	return SDL_GetTicks() - aTimer;
}



bool CApplication::MouseButtonDown(bool aRight)
{
	Assert(mTimerCallbackActive)
	return aRight ? mMouseRight : mMouseLeft;
}



bool CApplication::MouseButtonClicked(bool aRight)
{
	Assert(mTimerCallbackActive)
	return aRight ? mMouseRight && !mMouseRightLast : mMouseLeft && !mMouseLeftLast;
}



bool CApplication::MouseButtonReleased(bool aRight)
{
	Assert(mTimerCallbackActive)
	return aRight ? !mMouseRight && mMouseRightLast : !mMouseLeft && mMouseLeftLast;
}



void CApplication::GetCursorPos(int& aCursorX, int& aCursorY)
{
	SDL_GetMouseState(&aCursorX, &aCursorY);
}



bool CApplication::IsKeyDown(EKey aKey)
{
	Uint8 *KeyState = SDL_GetKeyState(NULL);
	return KeyState[GetSDLKey(aKey)] != 0;
}



void CApplication::OnKeyDown(SDLKey aVirtualKey)
{
	EKey Key;
	if (GetKey(aVirtualKey, Key))
		mMainController->OnKeyDown(Key);
}



void CApplication::OnKeyUp(SDLKey aVirtualKey)
{
	EKey Key;
	if (GetKey(aVirtualKey, Key))
		mMainController->OnKeyUp(Key);
}



bool CApplication::GetKey(SDLKey aVirtualKey, EKey& aKey)
{
	switch (aVirtualKey)
	{
	case SDLK_LEFT:
		aKey = kLeft;
		return true;
	case SDLK_RIGHT:
		aKey = kRight;
		return true;
	case SDLK_UP:
		aKey = kUp;
		return true;
	case SDLK_DOWN:
		aKey = kDown;
		return true;
	case SDLK_RETURN:
		aKey = kEnter;
		return true;
	case SDLK_SPACE:
		aKey = kSpace;
		return true;
	case SDLK_ESCAPE:
		aKey = kEscape;
		return true;
	case SDLK_BACKSPACE:
		aKey = kBackspace;
		return true;
	case SDLK_TAB:
		aKey = kTab;
		return true;
	case SDLK_LSHIFT:
		aKey = kLeftShift;
		return true;
	case SDLK_RSHIFT:
		aKey = kRightShift;
		return true;
	case SDLK_LCTRL:
		aKey = kLeftControl;
		return true;
	case SDLK_RCTRL:
		aKey = kRightControl;
		return true;
	case SDLK_LALT:
		aKey = kLeftAlt;
		return true;
	case SDLK_RALT:
		aKey = kRightAlt;
		return true;
	case SDLK_a:
		aKey = kA;
		return true;
	case SDLK_b:
		aKey = kB;
		return true;
	case SDLK_c:
		aKey = kC;
		return true;
	case SDLK_d:
		aKey = kD;
		return true;
	case SDLK_e:
		aKey = kE;
		return true;
	case SDLK_f:
		aKey = kF;
		return true;
	case SDLK_g:
		aKey = kG;
		return true;
	case SDLK_h:
		aKey = kH;
		return true;
	case SDLK_i:
		aKey = kI;
		return true;
	case SDLK_j:
		aKey = kJ;
		return true;
	case SDLK_k:
		aKey = kK;
		return true;
	case SDLK_l:
		aKey = kL;
		return true;
	case SDLK_m:
		aKey = kM;
		return true;
	case SDLK_n:
		aKey = kN;
		return true;
	case SDLK_o:
		aKey = kO;
		return true;
	case SDLK_p:
		aKey = kP;
		return true;
	case SDLK_q:
		aKey = kQ;
		return true;
	case SDLK_r:
		aKey = kR;
		return true;
	case SDLK_s:
		aKey = kS;
		return true;
	case SDLK_t:
		aKey = kT;
		return true;
	case SDLK_u:
		aKey = kU;
		return true;
	case SDLK_v:
		aKey = kV;
		return true;
	case SDLK_w:
		aKey = kW;
		return true;
	case SDLK_x:
		aKey = kX;
		return true;
	case SDLK_y:
		aKey = kY;
		return true;
	case SDLK_z:
		aKey = kZ;
		return true;
	case SDLK_0:
		aKey = k0;
		return true;
	case SDLK_1:
		aKey = k1;
		return true;
	case SDLK_2:
		aKey = k2;
		return true;
	case SDLK_3:
		aKey = k3;
		return true;
	case SDLK_4:
		aKey = k4;
		return true;
	case SDLK_5:
		aKey = k5;
		return true;
	case SDLK_6:
		aKey = k6;
		return true;
	case SDLK_7:
		aKey = k7;
		return true;
	case SDLK_8:
		aKey = k8;
		return true;
	case SDLK_9:
		aKey = k9;
		return true;
	default:
		return false;
	}
}



SDLKey CApplication::GetSDLKey(EKey aKey)
{
	switch (aKey)
	{
	case kLeft:
		return SDLK_LEFT;
	case kRight:
		return SDLK_RIGHT;
	case kUp:
		return SDLK_UP;
	case kDown:
		return SDLK_DOWN;
	case kEnter:
		return SDLK_RETURN;
	case kSpace:
		return SDLK_SPACE;
	case kEscape:
		return SDLK_ESCAPE;
	case kBackspace:
		return SDLK_BACKSPACE;
	case kTab:
		return SDLK_TAB;
	case kLeftShift:
		return SDLK_LSHIFT;
	case kRightShift:
		return SDLK_RSHIFT;
	case kLeftControl:
		return SDLK_LCTRL;
	case kRightControl:
		return SDLK_RCTRL;
	case kLeftAlt:
		return SDLK_LALT;
	case kRightAlt:
		return SDLK_RALT;
	case kA:
		return SDLK_a;
	case kB:
		return SDLK_b;
	case kC:
		return SDLK_c;
	case kD:
		return SDLK_d;
	case kE:
		return SDLK_e;
	case kF:
		return SDLK_f;
	case kG:
		return SDLK_g;
	case kH:
		return SDLK_h;
	case kI:
		return SDLK_i;
	case kJ:
		return SDLK_j;
	case kK:
		return SDLK_k;
	case kL:
		return SDLK_l;
	case kM:
		return SDLK_m;
	case kN:
		return SDLK_n;
	case kO:
		return SDLK_o;
	case kP:
		return SDLK_p;
	case kQ:
		return SDLK_q;
	case kR:
		return SDLK_r;
	case kS:
		return SDLK_s;
	case kT:
		return SDLK_t;
	case kU:
		return SDLK_u;
	case kV:
		return SDLK_v;
	case kW:
		return SDLK_w;
	case kX:
		return SDLK_x;
	case kY:
		return SDLK_y;
	case kZ:
		return SDLK_z;
	case k0:
		return SDLK_0;
	case k1:
		return SDLK_1;
	case k2:
		return SDLK_2;
	case k3:
		return SDLK_3;
	case k4:
		return SDLK_4;
	case k5:
		return SDLK_5;
	case k6:
		return SDLK_6;
	case k7:
		return SDLK_7;
	case k8:
		return SDLK_8;
	case k9:
		return SDLK_9;
	}
	throw CException("Invalid key");
}



double CApplication::Random()
{
	return (double)rand() / RAND_MAX;
}



int CApplication::Random(int aMin, int aMax)
{
	Assert(aMin <= aMax)
	if (aMin == aMax)
		return aMin;
	return rand() % (aMax - aMin + 1) + aMin;
}



void CApplication::AttachRenderingContext()
{
	#ifndef NO_THREADS
		Assert(glXGetCurrentContext() == NULL && !mAltAttached)
		bool MCResult = glXMakeCurrent(mDisplay, mWindow, mAltRenderingContext);
		Assert(MCResult)
	#else
		Assert(!mAltAttached)
	#endif
	mAltAttached = true;
}



void CApplication::LoadTexture(CString aName, bool aTransparent)
{
	CString TextureFilePath = GetDataFilePath(aName);
	SDL_Surface *Image = IMG_Load(TextureFilePath.GetData());
	if (Image == NULL)
		throw CObeliskException(SL("Image file not found: ") + aName);
	int XSize = Image->w;
	int XSize2 = NextPowerOf2(XSize);
	int YSize = Image->h;
	int YSize2 = NextPowerOf2(YSize);
	unsigned int *Data = new unsigned int [XSize2 * YSize2];
	memset(Data, 0, XSize2 * YSize2 * sizeof(unsigned int));
	SDL_PixelFormat *Format = Image->format;
	SDL_LockSurface(Image);
	for (int y = 0; y < YSize; y++)
	{
		for (int x = 0; x < XSize; x++)
		{
			Uint32 Pixel = *(Uint32*)((TByte*)Image->pixels + (y * Image->pitch + x * Format->BytesPerPixel));
			TByte R = ((Pixel & Format->Rmask) >> Format->Rshift) << Format->Rloss;
			TByte G = ((Pixel & Format->Gmask) >> Format->Gshift) << Format->Gloss;
			TByte B = ((Pixel & Format->Bmask) >> Format->Bshift) << Format->Bloss;
			unsigned int Colour = R | (G << 8) | (B << 16);
			if (aTransparent && Colour == 0xffffff)   // white: transparent
				Data[y * XSize2 + x] = 0x00000000;
			else if (aTransparent && Colour == 0xff00ff)   // magenta: semi-transparent
				Data[y * XSize2 + x] = 0x7f7f7f7f;
			else
				Data[y * XSize2 + x] = 0xff000000 | Colour;
		}
	}
	SDL_UnlockSurface(Image);
	SDL_FreeSurface(Image);
	GLuint Name;
	glGenTextures(1, &Name);
	glBindTexture(GL_TEXTURE_2D, Name);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, XSize2, YSize2, 0, GL_RGBA, GL_UNSIGNED_BYTE,
			Data);
	Assert(glGetError() == GL_NO_ERROR)
	mTextures.Add(aName, CTexture(XSize, YSize, (float)XSize / XSize2, (float)YSize / YSize2, Name));
}



int CApplication::NextPowerOf2(int n)
{
	int Next = 1;
	while (n > Next)
		Next <<= 1;
	return Next;
}



CTexture CApplication::GetTexture(CString aName)
{
	return mTextures[aName];
}



void CApplication::SetLinearFiltering(bool aLinearFiltering)
{
	mLinearFiltering = aLinearFiltering;
}



void CApplication::SetBackgroundColour(float aR, float aG, float aB)
{
	Assert(aR >= 0 && aR <= 1 && aG >= 0 && aG <= 1 && aB >= 0 && aB <= 1)
	glClearColor(aR, aG, aB, 1.0f);
}



void CApplication::BeginFrame()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	Assert(glGetError() == GL_NO_ERROR)
}



void CApplication::EndFrame()
{
	glXSwapBuffers(mDisplay, mWindow);
	Assert(glGetError() == GL_NO_ERROR)
}



void CApplication::SetSpriteViewport(int aXSize, int aYSize)
{
	GLenum MatrixMode;
	glGetIntegerv(GL_MATRIX_MODE, (GLint*)&MatrixMode);
	glMatrixMode(GL_PROJECTION);
	glTranslatef(-1, 1, 0);
	glScalef(2.0 / aXSize, -2.0 / aYSize, 1);
	glMatrixMode(MatrixMode);
}



void CApplication::DrawSprite(CTexture aTexture, float aX, float aY, float aXSize,
		float aYSize, double aXAlign, double aYAlign)
{
	int XSize = aXSize == -1 ? aTexture.GetXSize() : aXSize;
	int YSize = aYSize == -1 ? aTexture.GetYSize() : aYSize;
	double X = ((double)aX - aXAlign * XSize);
	double Y = ((double)aY - aYAlign * YSize);

	glEnable(GL_TEXTURE_2D);
	glDisable(GL_CULL_FACE);
	glBindTexture(GL_TEXTURE_2D, aTexture.GetTextureName());
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, mLinearFiltering ? GL_LINEAR : GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, mLinearFiltering ? GL_LINEAR : GL_NEAREST);
	glBegin(GL_QUADS);
	glTexCoord2f(0, 0);
	glColor3f(1, 1, 1);
	glVertex2f(X, Y);
	glTexCoord2f(aTexture.GetXRatio(), 0);
	glColor3f(1, 1, 1);
	glVertex2f(X + XSize, Y);
	glTexCoord2f(aTexture.GetXRatio(), aTexture.GetYRatio());
	glColor3f(1, 1, 1);
	glVertex2f(X + XSize, Y + YSize);
	glTexCoord2f(0, aTexture.GetYRatio());
	glColor3f(1, 1, 1);
	glVertex2f(X, Y + YSize);
	glEnd();
	Assert(glGetError() == GL_NO_ERROR)
}



void CApplication::TriColour(const CVector& aV0, const CVector& aV1, const CVector& aV2,
		const CColour& aColour)
{
	TriColour(aV0, aColour, aV1, aColour, aV2, aColour);
}

void CApplication::TriColour(const CVector& aV0, const CColour& aC0, const CVector& aV1,
		const CColour& aC1, const CVector& aV2, const CColour& aC2)
{
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_CULL_FACE);
	glBegin(GL_TRIANGLES);
	glColor3f(aC0.R, aC0.G, aC0.B);
	glVertex3f(aV0.X, aV0.Y, aV0.Z);
	glColor3f(aC1.R, aC1.G, aC1.B);
	glVertex3f(aV1.X, aV1.Y, aV1.Z);
	glColor3f(aC2.R, aC2.G, aC2.B);
	glVertex3f(aV2.X, aV2.Y, aV2.Z);
	glEnd();
	Assert(glGetError() == GL_NO_ERROR)
}



void CApplication::TriTexture(const CVector& aV0, float aS0, float aT0,
		const CVector& aV1, float aS1, float aT1,
		const CVector& aV2, float aS2, float aT2, CTexture aTexture)
{
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, aTexture.GetTextureName());
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, mLinearFiltering ? GL_LINEAR : GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, mLinearFiltering ? GL_LINEAR : GL_NEAREST);
	glEnable(GL_CULL_FACE);
	glBegin(GL_TRIANGLES);
	glTexCoord2f(aS0 * aTexture.GetXRatio(), aT0 * aTexture.GetYRatio());
	glColor3f(1, 1, 1);
	glVertex3f(aV0.X, aV0.Y, aV0.Z);
	glTexCoord2f(aS1 * aTexture.GetXRatio(), aT1 * aTexture.GetYRatio());
	glColor3f(1, 1, 1);
	glVertex3f(aV1.X, aV1.Y, aV1.Z);
	glTexCoord2f(aS2 * aTexture.GetXRatio(), aT2 * aTexture.GetYRatio());
	glColor3f(1, 1, 1);
	glVertex3f(aV2.X, aV2.Y, aV2.Z);
	glEnd();
	Assert(glGetError() == GL_NO_ERROR)
}



void CApplication::DrawText(CString aText, int aX, int aY)
{
	int X = aX;
	for (int i = 0; aText.HasNext(i); aText.Next(i))
	{
		CTexture& Texture = mFont[aText.GetChar(i)];
		DrawSprite(Texture, X, aY);
		X += Texture.GetXSize();
	}
}



void CApplication::LoadSound(CString aName)
{
	CString FilePath = GetDataFilePath(aName);
	SDL_AudioSpec Spec;
	TByte *Data;
	unsigned int Size;
	if (SDL_LoadWAV(FilePath.GetData(), &Spec, &Data, &Size) == NULL)
		throw CObeliskException(SL("Could not load sound file: ") + aName);
	mSounds.Add(aName, SSoundInfo(Data, Size));
}



void CApplication::Sound(CString aName)
{
	if (!mAudioInitialised)
	{
		mAudioInitialised = true;
		mAudioSpec.freq = 22050;
		mAudioSpec.format = AUDIO_S16SYS;
		mAudioSpec.channels = 2;
		mAudioSpec.samples = 2048;
		mAudioSpec.callback = AudioCallbackProc;
		mAudioSpec.userdata = this;
		if (SDL_OpenAudio(&mAudioSpec, NULL) != 0)
			throw CObeliskException(SL("Failed to open audio device"));
	}
	SDL_LockAudio();
	mCurrentSound = mSounds[aName];
	mCurrentSoundOffset = 0;
	SDL_UnlockAudio();
	SDL_PauseAudio(0);
}



void CApplication::AudioCallbackProc(void* userdata, Uint8* stream, int len)
{
	((CApplication*)userdata)->AudioCallback(stream, len);
}



void CApplication::AudioCallback(TByte* aBuffer, int aSize)
{
	if (mCurrentSoundOffset == mCurrentSound.Size)
	{
		SDL_PauseAudio(1);
		return;
	}
	int Size = Min(aSize, mCurrentSound.Size - mCurrentSoundOffset);
	mCurrentSoundOffset += Size;
	memcpy(aBuffer, mCurrentSound.Data, Size);
}



IReader* CApplication::OpenDataFile(CString aFileName)
{
	CString DataFilePath = GetDataFilePath(aFileName);
	FILE *File = fopen(DataFilePath.GetData(), "r");
	if (File == NULL)
		throw CObeliskException(SL("Data file not found: ") + aFileName);
	return new CFileWrapper(File);
}



IReader* CApplication::ReadUserFile(CString aFileName)
{
	CString FilePath;
	if (!GetUserFilePath(aFileName, false, FilePath))
		return NULL;
	FILE *File = fopen(FilePath.GetData(), "r");
	if (File == NULL)
		return NULL;
	return new CFileWrapper(File);
}



IWriter* CApplication::WriteUserFile(CString aFileName)
{
	CString FilePath;
	if (!GetUserFilePath(aFileName, true, FilePath))
		return NULL;
	FILE *File = fopen(FilePath.GetData(), "w");
	if (File == NULL)
		return NULL;
	return new CFileWrapper(File);
}
